"""
This module provides validation functions.
"""
from .validation import TypeViolation, AbstractType, IntType, BoolType, StringType, ListType, integer, string, boolean, double

__all__ = [
    "TypeViolation",
    "AbstractType",
    "IntType",
    "StringType",
    "BoolType",
    "ListType",

    "string",
    "boolean",
    "integer",
    "double",
]





"""
Logging utility class
"""
import logging
import sys
from typing import Optional, Dict

class Logger:
    """just syntactic sugar"""

    @classmethod
    def configure(cls,
                 default_level: int = logging.INFO,
                 format: str = "[%(asctime)s] %(levelname)s in %(filename)s:%(lineno)d - %(message)s",
                 levels: Optional[Dict[str, int]] = None, stream=sys.stdout):
        logging.basicConfig(level=default_level, format=format)
        if levels is not None:
            for name, level in levels.items():
                logging.getLogger(name).setLevel(level)



"""
This module provides utility functions.
"""
from .stringbuilder import StringBuilder
from .logger import Logger
from .serialization import TypeSerializer, TypeDeserializer, get_serializer, get_deserializer
from .copy_on_write_cache import CopyOnWriteCache

__all__ = [
    "StringBuilder",
    "Logger",

    "CopyOnWriteCache",

    "TypeSerializer",
    "TypeDeserializer",
    "get_serializer",
    "get_deserializer"
]



"""
deserialization functions
"""
from dataclasses import is_dataclass, fields
from functools import lru_cache
from typing import get_origin, get_args, Union

from pydantic import BaseModel

class TypeDeserializer:
    # constructor

    def __init__(self, typ):
        self.typ = typ
        self.deserializer = self._build_deserializer(typ)

    def __call__(self, value):
        return self.deserializer(value)

    # internal

    def _build_deserializer(self, typ):
        origin = get_origin(typ)
        args = get_args(typ)

        if origin is Union:
            # Optional[X] => Union[X, NoneType]
            deserializers = [self._build_deserializer(arg) for arg in args if arg is not type(None)]

            def deser_union(value):
                if value is None:
                    return None
                for d in deserializers:
                    try:
                        return d(value)
                    except Exception:
                        continue
                raise ValueError(f"Cannot deserialize value: {value!r} into Union[{args}]")

            return deser_union

        if isinstance(typ, type) and issubclass(typ, BaseModel):
            field_deserializers = {
                name: self._build_deserializer(field.annotation)
                for name, field in typ.model_fields.items()
            }

            def deser_model(value):
                if isinstance(value, typ):
                    return value
                if not isinstance(value, dict):
                    raise TypeError(f"Expected dict to construct {typ.__name__}, got {type(value).__name__}")
                kwargs = {
                    k: field_deserializers[k](v)
                    for k, v in value.items()
                    if k in field_deserializers
                }
                return typ.model_construct(**kwargs)

            return deser_model

        if is_dataclass(typ):
            field_deserializers = {
                f.name: self._build_deserializer(f.type) for f in fields(typ)
            }

            def deser_dataclass(value):
                if isinstance(value, typ):
                    return value
                if not isinstance(value, dict):
                    raise TypeError(f"Expected dict to construct {typ}, got {type(value).__name__}")
                return typ(**{
                    k: field_deserializers[k](v) for k, v in value.items() if k in field_deserializers
                })

            return deser_dataclass

        if origin is list:
            item_type = args[0] if args else Any
            item_deser = self._build_deserializer(item_type)
            return lambda v: [item_deser(item) for item in v]

        if origin is dict:
            key_type = args[0] if args else Any
            val_type = args[1] if len(args) > 1 else Any
            key_deser = self._build_deserializer(key_type)
            val_deser = self._build_deserializer(val_type)
            return lambda v: {key_deser(k): val_deser(val) for k, val in v.items()}

        # Fallback: primitive types, str, int, etc.
        return lambda v: typ(v) if callable(typ) else v


class TypeSerializer:
    def __init__(self, typ):
        self.typ = typ
        self.serializer = self._build_serializer(typ)

    def __call__(self, value):
        return self.serializer(value)

    def _build_serializer(self, typ):
        origin = get_origin(typ)
        args = get_args(typ)

        if origin is Union:
            serializers = [TypeSerializer(arg) for arg in args if arg is not type(None)]
            def ser_union(value):
                if value is None:
                    return None
                for s in serializers:
                    try:
                        return s(value)
                    except Exception:
                        continue
                return value
            return ser_union

        if isinstance(typ, type) and issubclass(typ, BaseModel):
            return lambda v: v.model_dump() if v is not None else None

        if is_dataclass(typ):
            field_serializers = {f.name: TypeSerializer(f.type) for f in fields(typ)}
            def ser_dataclass(obj):
                if obj is None:
                    return None
                return {k: field_serializers[k](getattr(obj, k)) for k in field_serializers}
            return ser_dataclass

        if origin is list:
            item_ser = TypeSerializer(args[0]) if args else lambda x: x
            return lambda v: [item_ser(item) for item in v] if v is not None else None

        if origin is dict:
            key_ser = TypeSerializer(args[0]) if args else lambda x: x
            val_ser = TypeSerializer(args[1]) if len(args) > 1 else lambda x: x
            return lambda v: {key_ser(k): val_ser(val) for k, val in v.items()} if v is not None else None

        # Fallback: primitive Typen oder unbekannt
        return lambda v: v

@lru_cache(maxsize=512)
def get_deserializer(typ) -> TypeDeserializer:
    """
    return a function that is able to deserialize a value of the specified type

    Args:
        typ: the type

    Returns:

    """
    return TypeDeserializer(typ)

@lru_cache(maxsize=512)
def get_serializer(typ) -> TypeSerializer:
    """
    return a function that is able to deserialize a value of the specified type

    Args:
        typ: the type

    Returns:

    """
    return TypeSerializer(typ)



"""
This module provides tools for mapping
"""
from .mapper import Mapper, MappingDefinition, matching_properties
from .transformer import Property, Operation
from. convert import Convert, Converter

__all__ = [
    # transformer

    "Property",
    "Operation",

    # mapper

    "Mapper",
    "MappingDefinition",
    "matching_properties"
]



"""
Test cases for the TypeDescriptor and Decorators functionality in aspyx.reflection.
"""
import unittest
from dataclasses import dataclass, fields

from pydantic import BaseModel

from aspyx.reflection import TypeDescriptor, Decorators


def transactional():
    def decorator(func):
        Decorators.add(func, transactional)
        return func #

    return decorator

@transactional()
class Base:
    def __init__(self):
        pass

    @transactional()
    def base(self, message: str) -> str:
        pass

    def no_type_hints(self, message):
        pass

class Derived(Base):
    @classmethod
    def foo(cls):
        pass

    def derived(self, message: str) -> str:
        pass

class Normal:
    def __init__(self, id: str):
        self.id = id

@dataclass
class Dataclass:
    id: str

class Pydantic(BaseModel):
    id: str

class TestReflection(unittest.TestCase):
    def test_properties(self):
        #normal_descriptor = TypeDescriptor.for_type(Normal)
        dataclass_descriptor = TypeDescriptor.for_type(Dataclass)
        #pydantic_descriptor = TypeDescriptor.for_type(Pydantic)

        fs = fields(Dataclass)

        print(1)

    def test_decorators(self):
        base_descriptor = TypeDescriptor.for_type(Base)

        self.assertTrue(base_descriptor.has_decorator(transactional))
        self.assertTrue( base_descriptor.get_method("base").has_decorator(transactional))

    def test_methods(self):
        derived_descriptor = TypeDescriptor.for_type(Derived)

        self.assertIsNotNone(derived_descriptor.get_method("derived").return_type, str)


if __name__ == '__main__':
    unittest.main()



import time
import unittest
from dataclasses import dataclass

from pydantic import BaseModel

from aspyx.mapper import Mapper, MappingDefinition, matching_properties, Convert
from aspyx.mapper.mapper import path
from aspyx.reflection import TypeDescriptor


class Class:
    id : str

    def __init__(self, id: str):
        self.id = id

@dataclass
class Money:
    currency : str
    value : int

@dataclass
class Product:
    name: str
    price: Money

@dataclass
class DataClass:
    id : str
    attr_1 : int
    attr_2 : int
    attr_3 : int
    attr_4 : int
    attr_5 : int
    attr_6 : int
    attr_7 : int
    attr_8 : int
    attr_9 : int

class Pydantic(BaseModel):
    id: str

@dataclass
class Types:
    b : bool
    i : int
    f : float
    s : str

@dataclass
class Deep:
    dc : DataClass
    dcs : list[DataClass]

def measure(n: int, func, name, *args, **kwargs):
    """
    Measure how long it takes to execute `func` n times.

    Args:
        n (int): number of iterations
        func (callable): the function to execute
        *args, **kwargs: arguments to pass to func
    """
    start = time.perf_counter()

    for _ in range(n):
        func(*args, **kwargs)

    end = time.perf_counter()

    total = end - start
    per_op = total / n if n > 0 else 0

    print(f"Total time for {name}: {total:.6f} seconds")
    print(f"Time per operation: {per_op * 1_000_000:.3f} µs ({per_op * 1000:.6f} ms)")


class TestMapper(unittest.TestCase):
    # instances

    c = Class(id="id")

    dc = DataClass(id="id", attr_1=1,  attr_2=2, attr_3=3, attr_4=4, attr_5=5, attr_6=6, attr_7=7, attr_8=8, attr_9=9)

    p = Pydantic(id="id")

    types = Types(b=False, i=1, f=1.0, s="s")

    deep = Deep(
        dc=dc,
        dcs = [dc, dc, dc, dc, dc, dc, dc, dc, dc, dc]
    )

    product = Product(name="p", price=Money(currency="EUR", value=1))

    def test_class(self):
        d = TypeDescriptor.for_type(Class)
        mapper = Mapper(
            MappingDefinition(source=Class, target=Class)
                .map(from_="id", to="id")
        )

        res = mapper.map(TestMapper.c)
        self.assertEqual(res.id, TestMapper.c.id)

    def test_data_class(self):
        mapper = Mapper(
            MappingDefinition(source=DataClass, target=DataClass)
                .map(all=matching_properties())
        )

        res = mapper.map(TestMapper.dc)
        self.assertEqual(res.id, TestMapper.dc.id)

    def test_pydantic(self):
        mapper = Mapper(
            MappingDefinition(source=Pydantic, target=Pydantic)
                .map(from_="id", to="id")
        )

        res = mapper.map(TestMapper.p)
        self.assertEqual(res.id, TestMapper.p.id)

    def test_finalize(self):
        finalized = False

        def mark_finalized():
            nonlocal finalized
            finalized = True

        mapper = Mapper(
            MappingDefinition(source=DataClass, target=DataClass)
            .map(all=matching_properties())
            .finalize(lambda source, target: mark_finalized())
        )

        mapper.map(TestMapper.dc)
        self.assertTrue(finalized)

    def test_deep(self):
        mapper = Mapper(
            MappingDefinition(source=Deep, target=Deep)
                .map(from_="dc", to="dc", deep=True)
                .map(from_="dcs", to="dcs", deep=True),

            MappingDefinition(source=DataClass, target=DataClass)
                .map(all=matching_properties())
        )

        res = mapper.map(TestMapper.deep)
        print(res)

    def test_wildcards(self):
        mapper = Mapper(
            MappingDefinition(source=Types, target=Types)
                .map(all=matching_properties())
        )

        res = mapper.map(TestMapper.types)
        #self.assertEqual(res.id, TestMapper.c.id)

    def test_path(self): # TODO
        d1 = TypeDescriptor.for_type(Product)
        d2 = TypeDescriptor.for_type(Money)

        mapper = Mapper(
            MappingDefinition(source=Product, target=Product)
            .map(from_="name", to="name")
            .map(from_=path("price", "currency"), to=path("price", "currency"))
            .map(from_=path("price", "value"), to=path("price", "value"))
        )

        res = mapper.map(TestMapper.product)
        print(res)


    def test_conversion(self):
        mapper = Mapper(
            MappingDefinition(source=Types, target=Types)
                .map(from_="b", to="b", convert=Convert(bool, bool, convert_source=lambda s: not s))
                .map(from_="i", to="f")
                .map(from_="f", to="i")
                .map(from_="s", to="s", convert=Convert(str, str, convert_source=lambda s: s+s))
        )

        res = mapper.map(TestMapper.types)
        print(res)
        #self.assertEqual(res.id, TestMapper.c.id)

    def test_benchmark(self):
        mapper = Mapper(
            MappingDefinition(source=Deep, target=Deep)
                .map(from_="dc", to="dc", deep=True)
                .map(from_="dcs", to="dcs", deep=True),

            MappingDefinition(source=DataClass, target=DataClass)
                .map(all=matching_properties())
        )

        # warm up

        mapper.map(TestMapper.deep)

        # benchmark

        loops = 100000

        def map_manual():
            deep = TestMapper.deep

            def copy_dc(o : DataClass):
                return DataClass(id=o.id,
                                 attr_1=o.attr_1,
                                 attr_2=o.attr_1,
                                 attr_3=o.attr_1,
                                 attr_4=o.attr_1,
                                 attr_5=o.attr_1,
                                 attr_6=o.attr_1,
                                 attr_7=o.attr_1,
                                 attr_8=o.attr_1,
                                 attr_9=o.attr_1)
            Deep(
                dc=copy_dc(deep.dc),
                dcs = [copy_dc(x) for x in deep.dcs]
            )

        def map_operation():
            mapper.map(TestMapper.deep)

        measure(loops, map_manual, "manual")
        measure(loops, map_operation, "mapper")


"""
This module provides a TypeDescriptor class that allows introspection of Python classes,
including their methods, decorators, and type hints. It supports caching for performance
"""
from __future__ import annotations

from abc import ABC, abstractmethod
import inspect
from inspect import signature
import threading
from types import FunctionType

from typing import Callable, get_type_hints, Type, Dict, Any, get_origin, List, get_args
from weakref import WeakKeyDictionary

from aspyx.validation import AbstractType, IntType
from aspyx.validation.validation import DoubleType, StringType, BoolType, ListType

import dataclasses
from dataclasses import MISSING, is_dataclass, fields
from typing import Any, Optional, Type
from pydantic import BaseModel, Field


try:
    from pydantic import Field
except ImportError:
    Field = None

def is_pydantic_model(cls: type) -> bool:
    return hasattr(cls, "__fields__")

def is_sqlalchemy_entity(cls: type) -> bool:
    # All mapped SQLAlchemy classes have a __mapper__ attribute
    return hasattr(cls, "__mapper__")

def is_dataclass(cls: type) -> bool:
    return hasattr(cls, "__dataclass_fields__")

def get_safe_type_hints(obj) -> Dict[str, Any]:
    """
    Safe wrapper around typing.get_type_hints that never raises.
    Returns either the resolved hints or a best-effort fallback (raw __annotations__).
    """
    try:
        # If obj is a function/method, use its globals so forward refs that are resolvable succeed
        if inspect.isfunction(obj) or inspect.ismethod(obj):
            return get_type_hints(obj, globalns=obj.__globals__, localns={})
        # classes / other objects
        return get_type_hints(obj)
    except Exception:
        # fallback: return raw annotations (may contain strings / TypeVar names)
        anns = getattr(obj, "__annotations__", {})
        # make a safe copy; don't attempt to eval strings
        return dict(anns)

def make_setter(cls: Type, field_name: str) -> Callable[[Any, Any], None]:
    attr = getattr(cls, field_name, None)

    # If it's a property with a fset, call that directly
    if isinstance(attr, property) and attr.fset:
        fset = attr.fset
        def setter(instance: Any, value: Any):
            fset(instance, value)
        return setter

    # Default: setattr
    def setter(instance: Any, value: Any):
        setattr(instance, field_name, value)

    return setter

def get_method_class(method):
    """
    return the class of the specified method
    Args:
        method: the method

    Returns:
        the class of the specified method

    """
    if inspect.ismethod(method) or inspect.isfunction(method):
        qualname = method.__qualname__
        module = inspect.getmodule(method)
        if module:
            cls_name = qualname.split('.<locals>', 1)[0].rsplit('.', 1)[0]
            cls = getattr(module, cls_name, None)
            if inspect.isclass(cls):
                return cls

    return None

def is_list_type(typ) -> bool:
    """
    Returns True if the given type hint represents a list-like container.
    Handles typing.List, list, and parameterized generics like list[int].
    """
    if typ is None:
        return False

    origin = get_origin(typ) or typ
    return origin in (list, List)

def get_list_element_type(typ) -> Any:
    """
    Returns the element type if `typ` is a list-like type.
    Examples:
        list[int]        -> int
        List[str]        -> str
        list[Any]        -> Any
        list             -> Any
        not a list       -> None
    """
    origin = get_origin(typ)
    if origin in (list, List):
        args = get_args(typ)
        return args[0] if args else Any
    return None

class DecoratorDescriptor:
    """
    A DecoratorDescriptor covers the decorator - a callable - and the passed arguments
    """
    __slots__ = [
        "decorator",
        "args"
    ]

    def __init__(self, decorator: Callable, *args):
        self.decorator = decorator
        self.args = args

    def __str__(self):
        return f"@{self.decorator.__name__}({', '.join(map(str, self.args))})"

class Decorators:
    """
    Utility class that caches decorators ( Python does not have a feature for this )
    """
    @classmethod
    def add(cls, func_or_class, decorator: Callable, *args):
        """
        Remember the decorator
        Args:
            func_or_class: a function or class
            decorator: the decorator
            *args: any arguments supplied to the decorator
        """
        current = func_or_class.__dict__.get('__decorators__')
        if current is None:
            setattr(func_or_class, '__decorators__', [DecoratorDescriptor(decorator, *args)])
        else:
            # Avoid mutating inherited list
            if '__decorators__' not in func_or_class.__dict__:
                current = list(current)
                setattr(func_or_class, '__decorators__', current)
            current.append(DecoratorDescriptor(decorator, *args))

    @classmethod
    def has_decorator(cls, func_or_class, callable: Callable) -> bool:
        """
        Return True, if the function or class is decorated with the decorator
        Args:
            func_or_class: a function or class
            callable: the decorator

        Returns:
            bool: the result
        """
        return any(decorator.decorator is callable for decorator in Decorators.get(func_or_class))

    @classmethod
    def get_decorator(cls, func_or_class, callable: Callable) -> DecoratorDescriptor:
        return next((decorator for decorator in Decorators.get_all(func_or_class) if decorator.decorator is callable), None)

    @classmethod
    def get_all(cls, func_or_class) -> list[DecoratorDescriptor]:
        return getattr(func_or_class, '__decorators__', [])

    @classmethod
    def get(cls, func_or_class) -> list[DecoratorDescriptor]:
        """
        return the list of decorators associated with the given function or class
        Args:
            func_or_class: the function or class

        Returns:
            list[DecoratorDescriptor]: the list
        """
        if inspect.ismethod(func_or_class):
            func_or_class = func_or_class.__func__  # unwrap bound method

        #return getattr(func_or_class, '__decorators__', []) #will return inherited as well
        return func_or_class.__dict__.get('__decorators__', [])


def attribute(*, primary_key: bool = False, type_property: Optional[Any] = None,
              default=MISSING, **extra):
    metadata = {"primary_key": primary_key, "type_property": type_property}
    metadata.update(extra)
    return dataclasses.field(default=default, metadata=metadata)

class PropertyExtractor(ABC):
    """Base interface for all property extraction strategies."""

    @abstractmethod
    def extract(self, cls: Type) -> Optional[Dict[str, "TypeDescriptor.PropertyDescriptor"]]:
        """
        Attempt to extract property descriptors for the given class.
        Return a dict if successful, or None if not applicable.
        """
        pass

class PydanticPropertyExtractor(PropertyExtractor):
    def extract(self, cls: Type):
        if not issubclass(cls, BaseModel):
            return None

        props = {}
        for name, field in cls.model_fields.items():
            typ = field.annotation
            default = field.default if field.default is not None else None

            # Pydantic metadata support
            extra = field.json_schema_extra or {}
            p = TypeDescriptor.PropertyDescriptor(cls, name, typ, default)
            p.primary_key = extra.get("primary_key", False)
            p.type_property = extra.get("type_property", None)
            props[name] = p

        return props
class DataclassPropertyExtractor(PropertyExtractor):
    def extract(self, cls: Type):
        if not is_dataclass(cls):
            return None

        type_hints = get_type_hints(cls)
        props = {}
        for f in fields(cls):
            meta = getattr(f, "metadata", {})
            typ = type_hints.get(f.name, f.type)  # <-- resolve "str" -> str

            default = None
            if f.default is not f.default_factory and f.default is not dataclasses.MISSING:
                default = f.default

            prop = TypeDescriptor.PropertyDescriptor(cls, f.name, typ, default)
            prop.primary_key = meta.get("primary_key", False)
            prop.type_property = meta.get("type_property", None)
            props[f.name] = prop

        return props

class DefaultPropertyExtractor(PropertyExtractor):
    def extract(self, cls: Type):
        try:
            hints = get_type_hints(cls)
        except Exception:
            hints = {}
        try:
            sig = inspect.signature(cls.__init__)
        except Exception:
            return None

        props = {}
        for name, param in sig.parameters.items():
            if name == "self":
                continue
            props[name] = TypeDescriptor.PropertyDescriptor(
                cls,
                name,
                hints.get(name, object),
                None if param.default is inspect.Parameter.empty else param.default
            )
        return props


class TypeDescriptor:
    """
    This class provides a way to introspect Python classes, their methods, decorators, and type hints.
    """

    # static

    _extractors: list[PropertyExtractor] = [
        PydanticPropertyExtractor(),
        DataclassPropertyExtractor(),
        DefaultPropertyExtractor()
    ]

    @classmethod
    def register_extractor(cls, extractor: PropertyExtractor):
        TypeDescriptor._extractors.insert(0, extractor)

    @classmethod
    def extract_properties(cls, type: Type) -> Optional[Dict[str, "TypeDescriptor.PropertyDescriptor"]]:
        for extractor in TypeDescriptor._extractors:
            properties = extractor.extract(type)
            if properties is not None:
                return properties

        raise Exception("no properties")

    # inner classes

    class ParameterDescriptor:
        def __init__(self, name: str, type: Type):
            self.name = name
            self.type = type

    class PropertyDescriptor:
        """
        Describes a class property (field) — can be read and written via reflection.
        """
        def __init__(self, cls: Type, name: str, typ: Optional[Type] = None, default: Any = None, type_property: Optional[AbstractType] = None):
            self.clazz = cls
            self.name = name
            self.type = typ or object
            self.default = default
            self.type_property = type_property
            if self.type_property is None:
                self.type_property = self._infer_type_property(self.type)

        def _infer_type_property(self, typ: Any):
            """
            Try to infer an AbstractType from Python type hints.
            This provides automatic mapping when no explicit type_property is set.
            """
            origin = get_origin(typ)

            if typ is int:
                return IntType()
            elif typ is float:
                return DoubleType()
            elif typ is str:
                return StringType()
            elif typ is bool:
                return BoolType()
            elif origin in (list, List):
                args = get_args(typ)
                elem_type = args[0] if args else Any
                element_spec = self._infer_type_property(elem_type)
                return ListType(element_spec)
            #TODO elif inspect.isclass(typ):
                # For complex/nested objects
            #    return ObjectType(typ)
            return None

        def get(self, instance):
            return getattr(instance, self.name, self.default)

        def set(self, instance, value):
            setattr(instance, self.name, value)

        def __str__(self):
            return f"Property({self.name}: {getattr(self.type, '__name__', self.type)})"

    class MethodDescriptor:
        """
        This class represents a method of a class, including its decorators, parameter types, and return type.
        """
        # constructor

        def __init__(self, cls, method: Callable):
            self.clazz = cls
            self.method = method
            self.decorators: list[DecoratorDescriptor] = Decorators.get(method)
            self.param_types : list[Type] = []
            self.params: list[TypeDescriptor.ParameterDescriptor] = []

            type_hints = get_safe_type_hints(method)
            sig = signature(method)

            for name, _ in sig.parameters.items():
                if name != 'self':
                    self.params.append(TypeDescriptor.ParameterDescriptor(name, type_hints.get(name)))
                    self.param_types.append(type_hints.get(name, object))

            self.return_type = type_hints.get('return', None)

        # public

        def get_name(self) -> str:
            """
            return the method name

            Returns:
                str: the method name
            """
            return self.method.__name__

        def get_doc(self, default = "") -> str:
            """
            return the method docstring

            Args:
                default: the default if no docstring is found

            Returns:
                str: the docstring
            """
            return self.method.__doc__ or default

        def is_async(self) -> bool:
            """
            return true if the method is asynchronous

            Returns:
                bool: async flag
            """
            return inspect.iscoroutinefunction(self.method)

        def get_decorators(self) -> list[DecoratorDescriptor]:
            return self.decorators

        def get_decorator(self, decorator: Callable) -> Optional[DecoratorDescriptor]:
            """
            return the DecoratorDescriptor - if any - associated with the passed Callable

            Args:
                decorator: the decorator

            Returns:
                Optional[DecoratorDescriptor]: the DecoratorDescriptor or None
            """
            for dec in self.decorators:
                if dec.decorator is decorator:
                    return dec

            return None

        def has_decorator(self, decorator: Callable) -> bool:
            """
            return True if the method is decorated with the decorator

            Args:
                decorator: the decorator callable

            Returns:
                bool: True if the method is decorated with the decorator
            """
            for dec in self.decorators:
                if dec.decorator is decorator:
                    return True

            return False

        def __str__(self):
            return f"Method({self.method.__name__})"

    # class properties

    _cache = WeakKeyDictionary()
    _lock = threading.RLock()

    # class methods

    @classmethod
    def for_type(cls, clazz: Type) -> TypeDescriptor:
        """
        Returns a TypeDescriptor for the given class, using a cache to avoid redundant introspection.
        """
        descriptor = cls._cache.get(clazz)
        if descriptor is None:
            with cls._lock:
                descriptor = cls._cache.get(clazz)
                if descriptor is None:
                    descriptor = TypeDescriptor(clazz)
                    cls._cache[clazz] = descriptor

        return descriptor

    # constructor

    def __init__(self, cls):
        self.cls = cls
        self.decorators = Decorators.get(cls)
        self.methods: Dict[str, TypeDescriptor.MethodDescriptor] = {}
        self.local_methods: Dict[str, TypeDescriptor.MethodDescriptor] = {}
        self.properties: Dict[str, TypeDescriptor.PropertyDescriptor] = {}

        # check superclasses

        self.super_types = [TypeDescriptor.for_type(x) for x in cls.__bases__ if not self._is_framework_class(x)]

        for super_type in self.super_types:
            self.methods = self.methods | super_type.methods

        # methods

        for name, member in self._get_local_members(cls):
            method = TypeDescriptor.MethodDescriptor(cls, member)
            self.local_methods[name] = method
            self.methods[name] = method

        # properties

        self.properties = TypeDescriptor.extract_properties(cls)

        # constructor

        self.constructor = self._create_constructor()

    # internal

    def _is_framework_class(self, cls):
        if cls is object:
            return True

        module = getattr(cls, "__module__", "")

        return module.startswith("pydantic.") or module.startswith("sqlalchemy.")

    def _create_constructor(self) -> Callable[..., object]:
        cls = self.cls

        def make(**kwargs: Any) -> object:
            return cls(**kwargs)

        def new():
                return cls.__new__(cls)

        #if is_dataclass(cls):# or is_sqlalchemy_entity(cls):
        #    return new

        return make

    def _get_local_members(self, cls):
        #return [
        #    (name, value)
        #    for name, value in getmembers(cls, predicate=inspect.isfunction)
        #    if name in cls.__dict__
        #]

        return [
            (name, attr)
            for name, attr in cls.__dict__.items()
            if isinstance(attr, FunctionType)
        ]

    # public new

    def constructor_parameters(self):
        return self.local_methods["__init__"].params

    def is_immutable(self) -> bool:
        return False

    def has_default_constructor(self) -> bool:
        #if is_sqlalchemy_entity(self.cls):
        #    return True

        #if is_dataclass(self.cls):
        #    return True

        return False

    # public

    def get_properties(self) -> list[TypeDescriptor.PropertyDescriptor]:
        return list(self.properties.values())

    def get_property_names(self) -> list[str]:
        return [prop.name for prop in self.properties.values()]

    def get_property(self, name: str) -> Optional[TypeDescriptor.PropertyDescriptor]:
        return self.properties.get(name)

    def has_property(self, name: str) -> bool:
        return self.properties.get(name) is not None

    def get_decorator(self, decorator: Callable) -> Optional[DecoratorDescriptor]:
        """
        Returns the first decorator of the given type, or None if not found.
        """
        for dec in self.decorators:
            if dec.decorator is decorator:
                return dec

        return None

    def has_decorator(self, decorator: Callable) -> bool:
        """
        Checks if the class has a decorator of the given type."""
        for dec in self.decorators:
            if dec.decorator is decorator:
                return True

        return False

    def get_methods(self, local = False) ->  list[TypeDescriptor.MethodDescriptor]:
        """
        Returns a list of MethodDescriptor objects for the class.
        If local is True, only returns methods defined in the class itself, otherwise includes inherited methods.
        """
        if local:
            return list(self.local_methods.values())
        else:
            return list(self.methods.values())

    def get_method(self, name: str, local = False) -> Optional[TypeDescriptor.MethodDescriptor]:
        """
        Returns a MethodDescriptor for the method with the given name.
        If local is True, only searches for methods defined in the class itself, otherwise includes inherited methods.
        """
        if local:
            return self.local_methods.get(name, None)
        else:
            return self.methods.get(name, None)



"""

"""
from abc import ABC, abstractmethod
from typing import Any, Callable, List, Optional, Type
from dataclasses import dataclass

from .convert import TypeConversions
from ..reflection.reflection import TypeDescriptor
from .transformer import Operation, Property

class MapperException(Exception):
    pass


class MapperProperty(Property["MappingContext"], ABC):
    """
    MapperProperty is a Property that can provide type information and optional validation.
    """

    @abstractmethod
    def get_type(self) -> type:
        """
        Return the type of the property.
        """
        pass

    def validate(self, value: Any) -> None:
        """
        Optional validation of the value. Default does nothing.
        Override in subclasses if validation is needed.
        """
        pass

class MapperPropertyd:
    def get(self, instance: Any, context: Any):
        raise NotImplementedError
    def set(self, instance: Any, value: Any, context: Any):
        raise NotImplementedError
    def get_type(self) -> Type:
        return object

class MapList2List(MapperProperty):
    def __init__(self, mapper, source_type: Type, target_type: Type, property: MapperProperty, factory: Callable):
        self.mapper = mapper
        self.source_type = source_type
        self.target_type = target_type
        self.property = property
        self.factory = factory
        self.mapping = None
        self.polymorphic = False
        #TODO if TypeDescriptor.has_type_static(source_type):
        self.polymorphic = False#bool(TypeDescriptor.for_type(source_type).child_classes()) TODO

    def set(self, instance, value, context):
        if value is None:
            return

        lst = list(value)
        result = self.factory()
        if self.polymorphic:
            for element in lst:
                mapping = self.mapper.get_source_mapping(type(element))
                result.append(self.mapper.map(element, context=context, mapping=mapping))
        else:
            if self.mapping is None:
                self.mapping = self.mapper.get_mapping_x(self.source_type, self.target_type)
            for element in lst:
                result.append(self.mapper.map(element, context=context, mapping=self.mapping))
        self.property.set(instance, result, context)

    def get(self, instance, context):
        return None

    def get_type(self):
        return self.property.get_type()

class MapDeep(MapperProperty):
    def __init__(self, mapper, source_type: Type, target_property: MapperProperty):
        self.mapper = mapper
        self.source_type = source_type
        self.target_property = target_property
        self.mapping = None
        self.polymorphic = False
        #TODOif TypeDescriptor.has_type_static(source_type):
        self.polymorphic = False#bool(TypeDescriptor.for_type(source_type).child_classes())

    def get(self, instance, context):
        return None

    def set(self, instance, value, context):
        if self.polymorphic:
            mapping = self.mapper.get_source_mapping(type(value))
            self.target_property.set(instance, self.mapper.map(value, context=context, mapping=mapping), context)
        else:
            if self.mapping is None:
                self.mapping = self.mapper.get_mapping_x(self.source_type, self.target_property.get_type())
            self.target_property.set(instance, self.mapper.map(value, context=context, mapping=self.mapping), context)

    def get_type(self):
        return self.target_property.get_type()

class ConvertProperty(MapperProperty):
    def __init__(self, property: MapperProperty, conversion: Callable):
        self.property = property
        self.conversion = conversion

    def get(self, instance, context):
        return self.conversion(self.property.get(instance, context))

    def set(self, instance, value, context):
        self.property.set(instance, self.conversion(value), context)

    def get_type(self):
        return object

class SetResultArgument(MapperProperty):
    def __init__(self, result_definition, index: int, property: MapperProperty):
        self.result_definition = result_definition
        self.index = index
        self.property = property
        self.result_definition.missing -= 1

    def get(self, instance, context):
        raise MapperException("wrong direction")

    def set(self, instance, value, context):
        if hasattr(self.property, 'validate'):
            try:
                self.property.validate(value)
            except Exception:
                pass
        context.get_result_buffer(self.result_definition.index).set(instance, value, self.property, self.index, context)

    def get_type(self):
        if self.index < self.result_definition.constructor_args:
            return self.result_definition.type_descriptor.constructor_parameters()[self.index].type
        else:
            return self.property.get_type()

class PeekValueProperty(MapperProperty):
    def __init__(self, index: int, property: MapperProperty):
        self.index = index
        self.property = property

    def get(self, instance, context):
        value = context.peek(self.index)
        if value is not None:
            return self.property.get(value, context)
        return None

    def set(self, instance, value, context):
        raise MapperException("not possible")

    def get_type(self):
        return self.property.get_type()

class PushValueProperty(MapperProperty):
    def __init__(self, index: int):
        self.index = index

    def get(self, instance, context):
        raise MapperException("not possible")

    def set(self, instance, value, context):
        context.push(value, self.index)

    def get_type(self):
        return object

# Value receivers

class ValueReceiver:
    def receive(self, context, instance, value):
        raise NotImplementedError

class SetPropertyValueReceiver(ValueReceiver):
    def __init__(self, property: MapperProperty):
        self.property = property
    def receive(self, context, instance, value):
        self.property.set(instance, value, context)

class SetResultPropertyValueReceiver(ValueReceiver):
    def __init__(self, result_index: int, index: int, property: Optional[MapperProperty]):
        self.result_index = result_index
        self.index = index
        self.property = property
    def receive(self, context, instance, value):
        context.get_result_buffer(self.result_index).set(instance, value, self.property, self.index, context)

class MappingResultValueReceiver(ValueReceiver):
    def receive(self, context, instance, value):
        if getattr(context, "current_state", None) is not None:
            context.current_state.result = value

# Source/Target trees, Buffer, IntermediateResultDefinition, OperationBuilder

class SourceNode:
    def __init__(self, accessor, match=None, parent=None):
        self.parent = parent
        self.accessor = accessor
        self.match = match
        self.children: List[SourceNode] = []
        self.stack_index = -1
        self.fetch_property: Optional[MapperProperty] = None
        self.type = accessor.type if hasattr(accessor, 'type') else object

    @property
    def is_root(self):
        return self.parent is None

    @property
    def is_leaf(self):
        return len(self.children) == 0

    def fetch_value(self, source_tree, expected_type, operations):
        if not self.is_root:
            self.parent.fetch_value(source_tree, expected_type, operations)

        if self.fetch_property is None:
            if self.is_root:
                self.fetch_property = self.accessor.make_transformer_property(False)
                self.type = self.accessor.type
            else:
                self.fetch_property = PeekValueProperty(index=self.parent.stack_index,
                                                        property=self.accessor.make_transformer_property(False))
                self.type = self.accessor.type
            if not self.is_leaf:
                self.stack_index = source_tree.stack_size
                source_tree.stack_size += 1
                operations.append(Operation(self.fetch_property, PushValueProperty(index=self.stack_index)))

    def insert_match(self, tree, match, index):
        root = next((c for c in self.children if c.accessor == match.paths[0][index]), None)
        if root is None:
            root = tree.make_node(self, match.paths[0][index], match if (len(match.paths[0]) - 1 == index) else None)
            self.children.append(root)
        if len(match.paths[0]) > index + 1:
            root.insert_match(tree, match, index + 1)

    def find_matching_node(self, match, index):
        if index < len(match.paths[0]):
            for child in self.children:
                if child.accessor == match.paths[0][index]:
                    return child.find_matching_node(match, index + 1)
        return self

class SourceTree:
    def __init__(self, typ, matches):
        self.roots: List[SourceNode] = []
        self.stack_size = 0
        self.type = typ
        for match in matches:
            self.insert_match(match)

    def insert_match(self, match):
        root = next((r for r in self.roots if r.accessor == match.paths[0][0]), None)
        if root is None:
            root = self.make_node(None, match.paths[0][0], match if len(match.paths[0]) == 1 else None)
            self.roots.append(root)
        if len(match.paths[0]) > 1:
            root.insert_match(self, match, 1)

    def find_node(self, match):
        for node in self.roots:
            if node.match == match:
                return node
            elif node.accessor == match.paths[0][0]:
                return node.find_matching_node(match, 1)
        return None

    def make_node(self, parent, accessor, match):
        accessor.resolve(parent.accessor.type if parent else self.type, False)
        return SourceNode(accessor=accessor, parent=parent, match=match)

class Buffer:
    def __init__(self, definition, n_args, constructor_args):
        constructor_args = n_args # TODO WTF
        ###
        self.definition = definition
        self.n_args = n_args
        self.constructor_args = constructor_args
        self.constructor = definition.constructor
        self.value_receiver = definition.value_receiver
        self.n_supplied_args = 0
        self.arguments = {}
        #self.array_arguments = [None] * constructor_args
        self.result = None
        if constructor_args == 0:
            self.result = self.constructor({})

    def set(self, instance, value, property, index, mapping_context):
        if self.n_supplied_args < self.constructor_args:
            #self.array_arguments[index] = value
            self.arguments[property.name] = value
            if self.n_supplied_args == self.constructor_args - 1:
                self.result = self.constructor(**self.arguments)
        else:
            property.set(self.result, value, mapping_context)
        self.n_supplied_args += 1
        if self.n_supplied_args == self.n_args:
            self.value_receiver.receive(mapping_context, instance, self.result)

class IntermediateResultDefinition:
    def __init__(self, type_descriptor: TypeDescriptor, constructor, index, n_args, value_receiver):
        self.type_descriptor = type_descriptor
        self.constructor = constructor
        self.index = index
        self.n_args = n_args
        self.value_receiver = value_receiver
        self.constructor_args = len(type_descriptor.properties)# TODO len(type_descriptor.constructor_parameters)
        self.missing = self.constructor_args

    def create_buffer(self):
        return Buffer(self, self.n_args, self.constructor_args)

type_conversions = TypeConversions()

class TargetNode:
    def __init__(self, accessor, match=None, parent=None):
        self.parent = parent
        self.accessor = accessor
        self.match = match
        self.children: List[TargetNode] = []
        self.stack_index = -1
        self.result_definition: Optional[IntermediateResultDefinition] = None
        self.fetch_property: Optional[MapperProperty] = None
        self.type = accessor.type

    @property
    def is_root(self):
        return self.parent is None

    @property
    def is_leaf(self):
        return len(self.children) == 0

    @property
    def is_inner_node(self):
        return len(self.children) > 0

    def compute_value_receiver(self):
        if self.parent and self.parent.result_definition is not None:
            self.parent.result_definition.missing -= 1
            return SetResultPropertyValueReceiver(
                result_index=self.parent.result_definition.index,
                index=self.accessor.index,
                property=self.accessor.make_transformer_property(True) #TODO if self.accessor.index >= self.parent.result_definition.constructor_args else None)
            )
        else:
            return SetPropertyValueReceiver(property=self.accessor.make_transformer_property(True))

    def try_convert(self, source_type, target_type):
        conv = type_conversions.get_converter(source_type, target_type)
        if conv is not None:
            return conv
        raise MapperException(f"cannot convert {source_type} to {target_type}")

    def calculate_conversion(self, source_node):
        conversion = self.match.operation.converter
        deep = self.match.operation.deep
        result = None
        source_type = source_node.accessor.type
        target_type = self.accessor.type
        if conversion is not None:
            from_type = conversion.source_type
            to_type = conversion.target_type
            if from_type != source_type:
                raise MapperException(f"conversion source type {from_type} does not match {source_type}")
            if to_type != target_type:
                raise MapperException(f"conversion target type {to_type} does not match {target_type}")
            result = conversion.source_converter()
        elif source_type != target_type and not deep:
            result = self.try_convert(source_type, target_type)
        return result

    def maybe_convert(self, prop: MapperProperty, conversion):
        if conversion is None:
            return prop
        return ConvertProperty(property=prop, conversion=conversion)

    def map_deep(self, mapper, source_accessor, target_accessor, target_property):
        is_source_multi = source_accessor.is_container()
        is_target_multi = target_accessor.is_container()
        if is_source_multi != is_target_multi:
            raise MapperException("relations must have the same cardinality")
        if is_source_multi:
            return MapList2List(mapper=mapper,
                                source_type=source_accessor.get_element_type(),
                                target_type=target_accessor.get_element_type(),
                                property=target_property,
                                factory=target_accessor.get_container_constructor())
        else:
            return MapDeep(mapper=mapper, source_type=source_accessor.type, target_property=target_property)

    def make_operation(self, source_node, mapper):
        source_property = source_node.fetch_property
        deep = self.match.operation.deep
        conversion = self.calculate_conversion(source_node)
        requires_write = self.parent and self.parent.result_definition is None
        write_property = self.accessor.make_transformer_property(requires_write)
        if self.parent and self.parent.result_definition is not None:
            self.parent.result_definition.missing -= 1
            write_property = SetResultArgument(self.parent.result_definition, self.accessor.index, write_property)
        if deep:
            write_property = self.map_deep(mapper, source_node.accessor, self.accessor, write_property)
        else:
            write_property = self.maybe_convert(write_property, conversion)
        return Operation(source_property, write_property)

    def make_operations(self, source_tree, target_tree, mapper, definition, operations):
        typ = self.accessor.type
        if self.is_root:
            descriptor = TypeDescriptor.for_type(target_tree.type)
            if descriptor.is_immutable() or not descriptor.has_default_constructor():
                self.result_definition = definition.add_intermediate_result_definition(TypeDescriptor.for_type(typ), descriptor.constructor, len(self.children), MappingResultValueReceiver())
            for child in self.children:
                child.make_operations(source_tree, target_tree, mapper, definition, operations)
        elif self.is_inner_node:
            descriptor = TypeDescriptor.for_type(typ)
            value_receiver = self.compute_value_receiver()
            constructor = descriptor.constructor
            self.result_definition = definition.add_intermediate_result_definition(descriptor, constructor, len(self.children), value_receiver)
            for child in self.children:
                child.make_operations(source_tree, target_tree, mapper, definition, operations)
        else:
            source_node = source_tree.find_node(self.match)
            source_node.fetch_value(source_tree, typ, operations)
            operations.append(self.make_operation(source_node, mapper))

    def insert_match(self, tree, match, index):
        root = next((c for c in self.children if c.accessor == match.paths[1][index]), None)
        if root is None:
            root = tree.make_node(self, match.paths[1][index], match if (len(match.paths[1]) - 1 == index) else None)
            self.children.append(root)
        if len(match.paths[1]) > index + 1:
            root.insert_match(tree, match, index + 1)

    def find_matching_node(self, match, index):
        if index < len(match.paths[0]):
            for child in self.children:
                if child.accessor == match.paths[0][index]:
                    return child.find_matching_node(match, index + 1)
        return self

class RootAccessor:
    def __init__(self, typ):
        self.name = ""
        self.type = typ
        self.index = 0
        self.read_only = False

    def make_transformer_property(self, write):
        raise NotImplementedError

    def resolve(self, typ, write):
        pass

    def __eq__(self, other):
        return isinstance(other, RootAccessor)

    def __hash__(self):
        return 1

class TargetTree:
    def __init__(self, typ, matches):
        self.root = TargetNode(accessor=RootAccessor(typ), parent=None, match=None)
        self.stack_size = 0
        self.type = typ
        for match in matches:
            self.root.insert_match(self, match, 0)

    def make_operations(self, source_tree, mapper, definition):
        operations = []
        self.root.make_operations(source_tree, self, mapper, definition, operations)
        return operations

    def make_node(self, parent, accessor, match):
        accessor.resolve(parent.accessor.type if parent else self.type, False)
        return TargetNode(accessor=accessor, parent=parent, match=match)

@dataclass
class OperationResult:
    operations: List[Operation]
    constructor: Callable
    stack_size: int

class OperationBuilder:
    def __init__(self, matches):
        self.matches = matches

    def make_operations(self, mapper, definition):
        source_tree = SourceTree(definition.source_class, self.matches)
        target_tree = TargetTree(definition.target_class, self.matches)
        operations = target_tree.make_operations(source_tree, mapper, definition)
        constructor = target_tree.root.result_definition.constructor if target_tree.root.result_definition else None

        type_descriptor = TypeDescriptor.for_type(definition.target_class)
        if constructor is None:
            constructor = type_descriptor.constructor#get_constructor()
        return OperationResult(operations=operations, constructor=constructor or (lambda: None), stack_size=source_tree.stack_size)



from dataclasses import dataclass
from lib2to3.btm_utils import pysyms

from pydantic import BaseModel

from aspyx.util import get_serializer, get_deserializer

@dataclass
class EmbeddedDataClass:
    name: str

embedded_dataclass = EmbeddedDataClass(name="foo")

class EmbeddedPydantic(BaseModel):
    name: str

embedded_pydantic = EmbeddedPydantic(name="foo")

@dataclass
class DataClass:
    int_attr : int
    bool_attr: bool
    int_list : list[int]

    embedded_dataclass : EmbeddedDataClass
    embedded_pydantic: EmbeddedPydantic

data_class = DataClass(int_attr=1, bool_attr=True, int_list=[1], embedded_pydantic=embedded_pydantic, embedded_dataclass=embedded_dataclass)

class Pydantic(BaseModel):
    int_attr: int
    bool_attr: bool

    int_list : list[int]

    embedded_dataclass: EmbeddedDataClass
    embedded_pydantic: EmbeddedPydantic

pydantic = Pydantic(int_attr=1, bool_attr=True,  int_list=[1], embedded_pydantic=embedded_pydantic, embedded_dataclass=embedded_dataclass)

class TestSerialization():
    def test_data_class(self):
        serializer = get_serializer(DataClass)
        deserializer = get_deserializer(DataClass)

        result = deserializer(serializer(data_class))

        assert data_class == result

    def test_pydantic(self):
        serializer = get_serializer(Pydantic)
        deserializer = get_deserializer(Pydantic)

        result = deserializer(serializer(pydantic))

        assert pydantic == result


from typing import TypeVar, Generic, Optional, Callable

K = TypeVar("K")
V = TypeVar("V")

class CopyOnWriteCache(Generic[K, V]):
    """
    cache, that clones the existing dict, whenever a new item is added, avoiding any locks
    """
    # constructor

    def __init__(self, factory: Optional[Callable[[K], V]] = None) -> None:
        self._cache: dict[K, V] = {}
        self._factory = factory

    # public

    def get(self, key: K, factory: Optional[Callable[[K], V]] = None) -> Optional[V]:
        value = self._cache.get(key, None)
        if value is None:
            if factory is not None:
                value = factory(key)
                self.put(key, value)

            elif self._factory is not None:
                value = self._factory(key)
                self.put(key, value)

        return value

    def put(self, key: K, value: V) -> None:
        new_cache = self._cache.copy()
        new_cache[key] = value
        self._cache = new_cache

    def contains(self, key: K) -> bool:
        return key in self._cache

    def clear(self) -> None:
        self._cache = {}


from dataclasses import dataclass
from typing import Optional, Type, Callable, Any, Dict, Tuple, TypeVar, Generic

Converter = Callable[[Any], Any]

S = TypeVar('S')
T = TypeVar('T')

class Convert(Generic[S, T]):
    def __init__(self, source_type: type[S], target_type: type[T], convert_source: Callable[[S], T], convert_target: Optional[Callable[[T], S]] = None):
        self.source_type = source_type
        self.target_type = target_type
        self.convert_source_func= convert_source
        self.convert_target_func= convert_target

    def convert_source(self, source: S) -> T:
        if self.convert_source_func is None:
            raise Exception("No convert_source function")

        return self.convert_source_func(source)

    def convert_target(self, source: T) -> S:
        if self.convert_target_func is None:
            raise Exception("No convert_target function")

        return self.convert_target_func(source)

    def source_converter(self):
        return lambda s: self.convert_source(s)

    def target_converter(self):
        return lambda s: self.convert_target(s)

class TypeConversions:
    def __init__(self):
        self._converters: Dict[Tuple[Type, Type], Converter] = {
            (str, int): lambda v: int(v) if v is not None else None,
            (str, float): lambda v: float(v) if v is not None else None,
            (str, bool): lambda v: str(v).lower() == 'true',
            (bool, int): lambda v: 1 if v else 0,
            (bool, float): lambda v: 1.0 if v else 0.0,
            (int, str): lambda v: str(v),
            (float, str): lambda v: str(v),
            (int, float): lambda v: float(v),
            (float, int): lambda v: int(v),
            (int, bool): lambda v: v == 1,
            (float, bool): lambda v: v == 1.0,
        }

    def register(self, convert: Convert):
        key = (convert.source_type, convert.target_type)
        self._converters[key] = convert.source_converter()

    def get_converter(self, from_type: Type, to_type: Type) -> Optional[Converter]:
        return self._converters.get((from_type, to_type))

    def convert(self, value, from_type: Type, to_type: Type):
        if value is None or from_type == to_type:
            return value
        converter = self._converters.get((from_type, to_type))
        if converter:
            return converter(value)
        if type(value) == to_type:
            return value
        raise TypeError(f"No converter registered for {from_type} -> {to_type}")


"""
context utility
"""
import contextvars
from contextlib import contextmanager
from typing import Generic, Optional, TypeVar, Any

T = TypeVar("T")

class ContextLocal(Generic[T]):
    """
    A context local value holder
    """
    # constructor

    def __init__(self, name: str, default: Optional[T] = None):
        self.var = contextvars.ContextVar(name, default=default)

    # public

    def get(self) -> Optional[T]:
        """
        return the current value or invoke the optional factory to compute one

        Returns:
            Optional[T]: the value associated with the current thread
        """
        return self.var.get()

    def set(self, value: Optional[T]) -> Any:
        """
        set a value in the current thread

        Args:
            value: the value

        Returns:
            a token that can be used as an argument to `reset`
        """
        return self.var.set(value)

    def reset(self, token) -> None:
        """
        clear the value in the current thread

        Args:
            token: the token to clear
        """
        self.var.reset(token)

    @contextmanager
    def use(self, value):
        token = self.set(value)
        try:
            yield
        finally:
            self.reset(token)



import unittest

from pydantic import BaseModel, Field
from pydantic.dataclasses import dataclass

from aspyx.reflection import TypeDescriptor
from aspyx.reflection.reflection import attribute
from aspyx.validation.validation import IntType, StringType, ValidationException, string


@dataclass
class DataClass:
    id : str = attribute(primary_key=True, type_property=string().max_length(10))

class PydanticClass(BaseModel):
    id: str = Field(json_schema_extra={"primary_key": True, "type_property": string().max_length(10)})

class TestValidation(unittest.TestCase):
    def test_constraints(self):
        type = IntType().min(5).max(10)

        self.assertTrue(type.is_valid(7))
        self.assertFalse(type.is_valid(3))


        type = StringType().not_empty().min_length(3)
        try:
            type.validate("hi")
            self.fail("should fail")
        except ValidationException as e:
            print("Validation failed:\n", e)

    def test_typedescriptor(self):
        data_class_descriptor = TypeDescriptor.for_type(DataClass)
        pydantic_class_descriptor =  TypeDescriptor.for_type(PydanticClass)

        self.assertTrue(True)


"""
A module with threading related utilities
"""
from .thread_local import ThreadLocal
from .context_local import ContextLocal

imports = [ThreadLocal]

__all__ = [
    "ThreadLocal",
    "ContextLocal",
]



from abc import ABC, abstractmethod
from typing import Any, Generic, List, TypeVar

CONTEXT = TypeVar("CONTEXT")

class Property(ABC, Generic[CONTEXT]):
    """
    A Property is able to read and write property values given an instance.

    CONTEXT: any context information that could be needed during a transformation
    """

    @abstractmethod
    def get(self, instance: Any, context: CONTEXT) -> Any:
        """Read a property value given an instance"""
        pass

    @abstractmethod
    def set(self, instance: Any, value: Any, context: CONTEXT) -> None:
        """Write a property value given an instance"""
        pass


class Operation(Generic[CONTEXT]):
    __slots__ = [
        "source",
        "target"
    ]

    """
    An Operation contains a source and a target Property that are used
    to read values from a source and set the result in the target object.
    """

    def __init__(self, source: Property[CONTEXT], target: Property[CONTEXT]):
        self.source = source
        self.target = target

    def set_target(self, from_: Any, to: Any, context: CONTEXT) -> None:
        """Set a target property by reading the appropriate value from the source"""
        value = self.source.get(from_, context)
        self.target.set(to, value, context)

    def set_source(self, to: Any, from_: Any, context: CONTEXT) -> None:
        """Set a source property by reading the appropriate value from the target"""
        value = self.target.get(to, context)
        self.source.set(from_, value, context)


class Transformer(Generic[CONTEXT]):
    """
    A Transformer is a generic class that transforms a source into a target object
    given a list of Operations.
    """

    __slots__ = [
        "operations"
    ]

    def __init__(self, operations: List[Operation[CONTEXT]]):
        self.operations = operations

    def transform_target(self, source: Any, target: Any, context: CONTEXT) -> None:
        for op in self.operations:
            op.set_target(source, target, context)



"""
Test cases for the TypeDescriptor and Decorators functionality in aspyx.reflection.
"""
from __future__ import annotations

import unittest

from aspyx.reflection import Decorators

def decorate_1(arg=""):
    def decorator(cls_or_func):
        Decorators.add(cls_or_func, decorate_1, arg)

        return cls_or_func #

    return decorator

def decorate_2(arg=""):
    def decorator(cls_or_func):
        Decorators.add(cls_or_func, decorate_2, arg)

        return cls_or_func #

    return decorator

@decorate_1("base")
class Base:
    def __init__(self):
        pass

    @decorate_1()
    def base(self, message: str) -> str:
        pass

    def no_type_hints(self, message):
        pass

@decorate_1("derived")
@decorate_2("derived")
class Derived(Base):
    @classmethod
    def foo(cls):
        pass

    def derived(self, message: str) -> str:
        pass

class OtherDerived(Base):
    pass

class TestReflection(unittest.TestCase):
    def test_decorators(self):
        base_decorators = Decorators.get(Base)
        derived_decorators = Decorators.get(Derived)
        other_derived_decorators = Decorators.get(OtherDerived)

        print()


if __name__ == '__main__':
    unittest.main()



from __future__ import annotations
from typing import Any, Callable, Type, TypeVar, Generic, Dict, List, Optional
from enum import Enum
import re

T = TypeVar('T')
B = TypeVar('B', bound='AbstractType')

# --- Helper Type Aliases ---
Check = Callable[[T], bool]
MethodApplier = Callable[['AbstractType', list[Any]], None]

# --- ArgType Enum ---
class ArgType(Enum):
    STRING = 'stringType'
    INT = 'intType'
    DOUBLE = 'doubleType'

    def parse(self, value: str) -> Any:
        if self is ArgType.STRING:
            return value
        elif self is ArgType.INT:
            return int(value)
        elif self is ArgType.DOUBLE:
            return float(value)

    @property
    def name(self) -> str:
        return self.value

# --- Test ---
class Test(Generic[T]):
    def __init__(self, *, type_: Type, name: str, check: Check[T],
                 params: Optional[Dict[str, Any]] = None,
                 stop: bool = False, message: Optional[str] = None):
        self.type = type_
        self.name = name
        self.check = check
        self.params = params or {}
        self.stop = stop
        self.message = message

    def run(self, obj: Any) -> bool:
        try:
            return self.check(obj)
        except Exception:
            return False

# --- MethodSpec ---
class MethodSpec:
    def __init__(self, arg_count: int, arg_types: List[ArgType], apply: MethodApplier):
        self.arg_count = arg_count
        self.arg_types = arg_types
        self.apply = apply

# --- Validation System ---
class TypeViolation:
    def __init__(self, *, type_: Type, name: str, params: Dict[str, Any],
                 value: Any, path: str, message: str):
        self.type = type_
        self.name = name
        self.params = params
        self.value = value
        self.path = path
        self.message = message

    def __str__(self):
        return f"[{self.path}] {self.name} failed on {self.value} ({self.message})"

class ValidationContext:
    def __init__(self):
        self.violations: List[TypeViolation] = []
        self.path: str = ""

    def add_violation(self, *, type_: Type, name: str, params: Dict[str, Any],
                      value: Any, path: str, message: str = ""):
        self.violations.append(TypeViolation(
            type_=type_, name=name, params=params, value=value,
            path=path, message=message
        ))

    @property
    def has_violations(self) -> bool:
        return bool(self.violations)

class ValidationException(Exception):
    def __init__(self, violations: List[TypeViolation]):
        self.violations = violations
        super().__init__(str(self))

    def __str__(self):
        return "\n".join(str(v) for v in self.violations)

# --- AbstractType Base ---
class AbstractType(Generic[T, B]):
    def __init__(self, *, type_: Type):
        self.type = type_
        self.nullable = False
        self.tests: List[Test[Any]] = []

    def base_type(self, type_: Type):
        self.type = type_
        self.test(
            type_=type_,
            name="type",
            params={"type": type_},
            check=lambda obj: (obj is None and self.nullable) or isinstance(obj, type_),
            stop=True
        )

    def test(self, *, type_: Type, name: str, check: Check[Any],
             params: Optional[Dict[str, Any]] = None,
             stop: bool = False, message: Optional[str] = None) -> B:
        self.tests.append(Test(type_=type_, name=name, check=check,
                               params=params, stop=stop, message=message))
        return self  # type: ignore

    def check_value(self, obj: Any, context: ValidationContext):
        for test in self.tests:
            if not test.run(obj):
                context.add_violation(
                    type_=test.type,
                    name=test.name,
                    params=test.params,
                    path=context.path,
                    value=obj,
                    message=test.message or ""
                )
                if test.stop:
                    break

    def validate(self, obj: Any):
        context = ValidationContext()
        self.check_value(obj, context)
        if context.has_violations:
            raise ValidationException(context.violations)

    def is_valid(self, obj: Any) -> bool:
        context = ValidationContext()
        self.check_value(obj, context)
        return not context.has_violations

    def optional(self) -> B:
        self.nullable = True
        if self.tests:
            self.tests[0].stop = True
        return self  # type: ignore

    def required(self) -> B:
        self.nullable = False
        return self  # type: ignore

# --- Concrete Implementations ---
class IntType(AbstractType[int, 'IntType']):
    def __init__(self):
        super().__init__(type_=int)
        self.base_type(int)

    def min(self, value: int) -> 'IntType':
        return self.test(type_=int, name="min", params={"min": value}, check=lambda x: x >= value)

    def max(self, value: int) -> 'IntType':
        return self.test(type_=int, name="max", params={"max": value}, check=lambda x: x <= value)

    def less_than(self, value: int) -> 'IntType':
        return self.test(type_=int, name="lessThan", params={"lessThan": value}, check=lambda x: x < value)

    def greater_than(self, value: int) -> 'IntType':
        return self.test(type_=int, name="greaterThan", params={"greaterThan": value}, check=lambda x: x > value)

class DoubleType(AbstractType[float, 'DoubleType']):
    def __init__(self):
        super().__init__(type_=float)
        self.base_type(float)

    def min(self, value: float) -> 'DoubleType':
        return self.test(type_=float, name="min", params={"min": value}, check=lambda x: x >= value)

    def max(self, value: float) -> 'DoubleType':
        return self.test(type_=float, name="max", params={"max": value}, check=lambda x: x <= value)

class StringType(AbstractType[str, 'StringType']):
    def __init__(self):
        super().__init__(type_=str)
        self.base_type(str)

    def not_empty(self) -> 'StringType':
        return self.test(type_=str, name="notEmpty", check=lambda s: bool(s and len(s) > 0))

    def min_length(self, length: int) -> 'StringType':
        return self.test(type_=str, name="minLength", params={"minLength": length}, check=lambda s: len(s) >= length)

    def max_length(self, length: int) -> 'StringType':
        return self.test(type_=str, name="maxLength", params={"maxLength": length}, check=lambda s: len(s) <= length)

    def re(self, pattern: str, message: Optional[str] = None) -> 'StringType':
        regex = re.compile(pattern)
        return self.test(type_=str, name="re", check=lambda s: bool(regex.match(s)), message=message)

class BoolType(AbstractType[bool, 'BoolType']):
    def __init__(self):
        super().__init__(type_=bool)
        self.base_type(bool)

# --- List Type ---
class ListType(AbstractType[list, 'ListType']):
    def __init__(self, element_type: AbstractType):
        super().__init__(type_=list)
        self.element_type = element_type
        self.test(type_=list, name="type", check=lambda obj: isinstance(obj, list), stop=True)

    def min(self, length: int) -> 'ListType':
        return self.test(type_=list, name="min", params={"min": length}, check=lambda s: len(s) >= length)

    def max(self, length: int) -> 'ListType':
        return self.test(type_=list, name="max", params={"max": length}, check=lambda s: len(s) <= length)


# some constants

def string():
    return StringType()

def boolean():
    return BoolType()

def integer():
    return IntType()

def double():
    return DoubleType()



"""

"""

from typing import Any, Callable, Dict, List, Optional, Tuple, Type
from dataclasses import dataclass

from .convert import Convert
from .operation_builder import MapperException, MapperProperty, IntermediateResultDefinition, OperationBuilder
from .transformer import Transformer
from ..reflection.reflection import TypeDescriptor, is_list_type, get_list_element_type


# Property implementations used by Accessors


def make_setter(cls: Type, field_name: str) -> Callable[[Any, Any], None]:
    attr = getattr(cls, field_name, None)

    # If it's a property with a fset, call that directly
    if isinstance(attr, property) and attr.fset:
        fset = attr.fset
        def setter(instance: Any, value: Any):
            fset(instance, value)
        return setter

    # Default: setattr
    def setter(instance: Any, value: Any):
        setattr(instance, field_name, value)

    return setter

class PropertyProperty:
    def __init__(self, field: TypeDescriptor.PropertyDescriptor, model_class=None):
        """
        :param field: A descriptor that should have a 'name' attribute.
        :param model_class: Optional Pydantic model class to check fields from.
        """
        self.field = field
        self.name = getattr(field, "name", None)
        self.getter = getattr(field, "getter", lambda obj: getattr(obj, self.name))
        self._setter_func = make_setter(field.clazz, field.name)
        self.model_class = model_class  # Needed to inspect Pydantic model_fields

    def get(self, instance, context=None):
        return self.getter(instance)

    def set(self, instance, value, context=None):
        result = self._setter_func(instance, value)
        # For frozen models, setter returns new instance, otherwise None
        return result if result is not None else instance

    def get_type(self):
        return getattr(self.field, "type", object)

class ValidatingPropertyProperty(PropertyProperty):
    def __init__(self, field):
        super().__init__(field)
        self.type_info = getattr(field, 'type', None)

    def set(self, instance, value, context):
        if hasattr(self.type_info, 'validate'):
            self.type_info.validate(value)
        super().set(instance, value, context)

    def get_type(self):
        return getattr(self.type_info, 'type', object)

class ConstantValue(MapperProperty):
    __slots__ = [
        "value"
    ]

    def __init__(self, value):
        self.value = value

    def get(self, instance, context):
        return self.value

    def set(self, instance, value, context):
        pass

    def get_type(self):
        return type(self.value)

# Accessor base classes

class Accessor:
    def __init__(self, name: str, typ: Type, index: int, read_only: bool=False):
        self.name = name
        self.index = index
        self.read_only = read_only
        self.type = typ
        # mapper will be set by context when needed
        self.mapper = None

    def resolve(self, typ: Type, write: bool):
        raise NotImplementedError

    def make_transformer_property(self, write: bool) -> MapperProperty:
        raise NotImplementedError

    def is_container(self):
        return is_list_type(self.type)

    def get_element_type(self):
        return get_list_element_type(self.type)

    def get_container_constructor(self):
        return list if self.is_container() else None # TODO just list now

class ConstantAccessor(Accessor):
    def __init__(self, value):
        super().__init__(name=str(value), typ=type(value), index=0, read_only=True)
        self.value = value

    def resolve(self, typ: Type, write: bool):
        if write:
            raise MapperException("constants are not writeable")

    def make_transformer_property(self, write: bool) -> MapperProperty:
        if write:
            raise MapperException("constants are not writeable")
        return ConstantValue(self.value)

    def __eq__(self, other):
        return isinstance(other, ConstantAccessor) and self.value == other.value

    def __hash__(self):
        return hash(self.value) if self.value is not None else 0

class PropertyAccessor(Accessor):
    def __init__(self, name: str, validate: bool=False):
        super().__init__(name=name, typ=object, index=-1, read_only=False)
        self.validate = validate
        self.field = None

    def make_transformer_property(self, write: bool):
        if write:
            if self.field is None:
                raise MapperException("field not resolved")
            if getattr(self.field, 'is_writeable', lambda: True)():
                return ValidatingPropertyProperty(self.field) if self.validate else PropertyProperty(self.field)
            else:
                raise MapperException(f"{self.field.type_descriptor.type}.{self.name} is final")
        else:
            return ValidatingPropertyProperty(self.field) if self.validate else PropertyProperty(self.field)

    #def get_container_constructor(self):
    #    return getattr(self.field, 'factory_constructor', None)

    #def get_element_type(self):
    #    return getattr(self.field, 'element_type', getattr(self.field, 'type', object))

    def resolve(self, typ: Type, write: bool):
        descriptor = TypeDescriptor.for_type(typ)
        init = descriptor.get_method("__init__")
        if init is not None:
            params = init.params
            idx = next((i for i, p in enumerate(params) if p.name == self.name), -1)
            self.index = idx
        else:
            self.index = -1

        self.field = descriptor.get_property(self.name)
        self.type = getattr(self.field, 'type', object)

    def __eq__(self, other):
        return isinstance(other, PropertyAccessor) and self.name == other.name

    def __hash__(self):
        return hash(self.name)

# MapOperation classes and qualifiers

class MapOperation:
    def __init__(self, converter: Optional[Convert]=None, deep: bool=False):
        self.converter = converter
        self.deep = deep
    def find_matches(self, definition, matches):
        raise NotImplementedError

class PropertyQualifier:
    def compute_properties(self, source_class: Type, target_class: Type) -> List[str]:
        raise NotImplementedError

class Properties(PropertyQualifier):
    def __init__(self, properties: List[str]):
        self.properties = properties
    def compute_properties(self, source_class: Type, target_class: Type) -> List[str]:
        return self.properties

class AllProperties(PropertyQualifier):
    def __init__(self):
        self.exceptions: List[str] = []
    def except_properties(self, properties: List[str]):
        self.exceptions.extend(properties)
        return self
    def compute_properties(self, source_class: Type, target_class: Type) -> List[str]:
        result = []
        src_desc = TypeDescriptor.for_type(source_class)
        tgt_desc = TypeDescriptor.for_type(target_class)
        names = src_desc.get_property_names()

        for prop in names:
            if prop in self.exceptions:
                continue
            if src_desc.has_property(prop) and tgt_desc.has_property(prop):
                result.append(prop)
        return result

def properties(properties_list: List[str]) -> Properties:
    return Properties(properties_list)

def matching_properties() -> AllProperties:
    return AllProperties()

@dataclass
class Match:
    operation: MapOperation
    source: List[Accessor]
    target: List[Accessor]
    paths: List[List[Accessor]] = None
    def __post_init__(self):
        self.paths = [self.source, self.target]

class MapProperties(MapOperation):
    def __init__(self, qualifier: PropertyQualifier, converter: Optional[Convert]=None, deep: bool=False):
        super().__init__(converter=converter, deep=deep)
        self.qualifier = qualifier
    def compute_properties(self, source_class: Type, target_class: Type) -> List[str]:
        return self.qualifier.compute_properties(source_class, target_class)
    def find_matches(self, definition, matches: List[Match]):
        for prop in self.compute_properties(definition.source_class, definition.target_class):
            matches.append(Match(operation=self, source=[PropertyAccessor(name=prop)], target=[PropertyAccessor(name=prop)]))

class MapAccessor(MapOperation):
    def __init__(self, source: List[Accessor], target: List[Accessor], converter: Optional[Convert]=None, deep: bool=False):
        super().__init__(converter=converter, deep=deep)
        self.source = source
        self.target = target
    def find_matches(self, definition, matches: List[Match]):
        matches.append(Match(operation=self, source=self.source, target=self.target))

# MappingDefinition, Mapping, MappingContext, Mapper

def path(*args: str):
    return [x for x in args if x is not None]

class MappingDefinition:
    def __init__(self, source: Type=None, target: Type=None):
        self.source_class = source
        self.target_class = target
        self.operations: List[MapOperation] = []
        self.intermediate_result_definitions = []
        self.finalizer = None
        self.base_mapping = None

    def add_intermediate_result_definition(self, type_descriptor, constructor, nargs, value_receiver):
        ir = IntermediateResultDefinition(type_descriptor=type_descriptor, constructor=constructor, index=len(self.intermediate_result_definitions), n_args=nargs, value_receiver=value_receiver)
        self.intermediate_result_definitions.append(ir)
        return ir

    def collect_finalizer(self):
        finalizers = []
        def collect(defn):
            if defn.base_mapping is not None:
                collect(defn.base_mapping)
            if defn.finalizer is not None:
                finalizers.append(defn.finalizer)
        collect(self)
        return finalizers

    def find_matches(self, matches: List[Match]):
        if self.base_mapping is not None:
            self.base_mapping.find_matches(matches)
        for op in self.operations:
            op.find_matches(self, matches)

    def create_operations(self, mapper):
        matches = []
        self.find_matches(matches)

        return OperationBuilder(matches).make_operations(mapper, self)

    def create_mapping(self, mapper):
        result = self.create_operations(mapper)
        for intermediate in self.intermediate_result_definitions:
            if intermediate.missing > 0:
                raise MapperException(f"{intermediate.type_descriptor.type} misses {intermediate.missing} arguments")
        mapping = Mapping(mapper=mapper,
                          definition=self,
                          constructor=result.constructor,
                          stack_size=result.stack_size,
                          intermediate_result_definitions=self.intermediate_result_definitions,
                          operations=result.operations,
                          finalizer=self.collect_finalizer())
        return mapping

    def map(self, constant=None, from_=None, all=None, to=None, deep=False, validate=False, convert: Optional[Convert]=None):
        if all is not None:
            self.operations.append(MapProperties(qualifier=all, converter=convert, deep=deep))
        else:
            from_accessors = []
            to_accessors = []
            if isinstance(from_, Accessor):
                from_accessors.append(from_)
            elif isinstance(from_, str):
                from_accessors.append(PropertyAccessor(name=from_))
            elif isinstance(from_, list):
                from_accessors = [PropertyAccessor(name=s) for s in from_]
            if constant is not None:
                from_accessors.append(ConstantAccessor(value=constant))
            if isinstance(to, Accessor):
                to_accessors.append(to)
            elif isinstance(to, str):
                to_accessors.append(PropertyAccessor(name=to, validate=validate))
            elif isinstance(to, list):
                to_accessors = [PropertyAccessor(name=s, validate=validate) for s in to]
            self.operations.append(MapAccessor(source=from_accessors, target=to_accessors, converter=convert, deep=deep))
        return self

    def finalize(self, finalizer_func: Callable):
        self.finalizer = finalizer_func
        return self

def mapping():
    return MappingDefinition()

class MappingState:
    def __init__(self, context):
        self.context = context
        self.result_buffers = context.result_buffers
        self.stack = context.stack
        self.result = None
        self.next_state = context.current_state
        context.current_state = self
    def restore(self, context):
        context.result_buffers = self.result_buffers
        context.stack = self.stack
        context.current_state = self.next_state

class MappingContext:
    def __init__(self, mapper):
        self.mapper = mapper
        self.current_source = None
        self.current_target = None
        self.mapped_objects = None
        self.result_buffers = []
        self.stack = []
        self.current_state = None
        if getattr(mapper, 'check_cycles', False):
            self.mapped_objects = {}
    def remember(self, source, target):
        if self.mapped_objects is not None:
            self.mapped_objects[id(source)] = target
        self.current_source = source
        self.current_target = target
        return self
    def mapped_object(self, source):
        return None if self.mapped_objects is None else self.mapped_objects.get(id(source))
    def setup_result_buffers(self, buffers):
        saved = self.result_buffers
        self.result_buffers = buffers
        return saved
    def setup(self, intermediate_result_definitions, stack_size: int):
        buffers = [ir.create_buffer() for ir in intermediate_result_definitions]
        if stack_size > 0:
            self.stack = [None] * stack_size
        return self.setup_result_buffers(buffers)
    def get_result_buffer(self, index: int):
        return self.result_buffers[index]
    def push(self, value, index: int):
        self.stack[index] = value
    def peek(self, index: int):
        return self.stack[index]

class MappingKey:
    def __init__(self, source: Type, target: Type):
        self.source = source
        self.target = target
    def _type_matches(self, a: Type, b: Type) -> bool:
        return a == b or a == object or b == object
    def __eq__(self, other):
        if not isinstance(other, MappingKey):
            return False
        return self._type_matches(self.source, other.source) and self._type_matches(self.target, other.target)
    def __hash__(self):
        return 0

class Mapping(Transformer[MappingContext]):
    def __init__(self, mapper, definition, constructor, stack_size, intermediate_result_definitions, operations, finalizer):
        super().__init__(operations=operations)

        self.mapper = mapper
        self.definition = definition
        self.constructor = constructor
        self.stack_size = stack_size
        self.intermediate_result_definitions = intermediate_result_definitions
        self.finalizer = finalizer or []
        self.type_descriptor = TypeDescriptor.for_type(definition.target_class)
        self.lazy = self.type_descriptor.is_immutable() or not self.type_descriptor.has_default_constructor()
    def setup_context(self, context: MappingContext):
        state = MappingState(context)
        context.setup(self.intermediate_result_definitions, self.stack_size)
        return state

    def new_instance(self):
        return self.constructor()
    def transform_target(self, source, target, context: MappingContext):
        for op in self.operations:
            op.set_target(source, target, context)

class Mapper:

    def __init__(self, *definitions: MappingDefinition, config: Optional[dict]=None):
        self.mapping_definitions = list(definitions)
        self.mappings: Dict[MappingKey, Mapping] = {}
        self.by_source_type_mappings: Dict[Type, Mapping] = {}
        self.check_cycles = False
        for definition in definitions:
            mapping = definition.create_mapping(self)
            self.register_mapping(mapping)
        if config is not None:
            self.check_cycles = config.get('check_cycles', False)
    def create_context(self):
        return MappingContext(self)
    def register_mapping(self, mapping: Mapping):
        key = MappingKey(mapping.definition.source_class, mapping.definition.target_class)
        self.mappings[key] = mapping
    def has_definition(self, source: Type, target: Type) -> bool:
        for d in self.mapping_definitions:
            if d.source_class == source and d.target_class == target:
                return True
        return False
    def get_source_mapping(self, source: Type):
        m = self.by_source_type_mappings.get(source)
        if m is None:
            for mapping in self.mappings.values():
                if mapping.definition.source_class == source:
                    self.by_source_type_mappings[source] = mapping
                    return mapping
            raise MapperException(f"no mapping for source type {source}")
        return m
    def get_mapping_x(self, source: Type, target: Type):
        key = MappingKey(source, target)
        m = self.mappings.get(key)
        if m is None:
            raise MapperException(f"No mapping found for <{source}, {target}>")
        return m

    def map(self, source, target: Optional[Any] = None, context: Optional[MappingContext]=None, mapping: Optional[Mapping]=None):
        if source is None:
            return None
        # if mapping not provided, try to infer by source runtime type and a registered mapping
        mapping = mapping or self.get_source_mapping(type(source))
        context = context or MappingContext(self)
        if target is None:
            target = context.mapped_object(source)
        lazy_create = False
        if target is None:
            lazy_create = mapping.lazy
            if lazy_create:
                target = context
            else:
                target = mapping.new_instance()
                context.remember(source, target)

        state = mapping.setup_context(context)
        try:
            mapping.transform_target(source, target, context)
            if lazy_create:
                target = context.current_state.result
                context.remember(source, target)
        finally:
            state.restore(context)

        for finalizer in mapping.finalizer:
            finalizer(source, target)

        return target
